from azureml.core import Workspace
from azureml.exceptions import WorkspaceException
from .exceptions import AzureException
from .decorators import error_handler
from .credentials import Credentials

class AzureProject(object):

    def __init__(self, ctx):
        super(AzureProject, self).__init__()
        self.ctx = ctx
        self.credentials = Credentials(self.ctx).load()
        self.credentials.verify()

    @error_handler    
    def list(self):
        workspaces = Workspace.list(
            self.credentials.subscription_id,
            auth=self.credentials.get_serviceprincipal_auth())
        for project in iter(workspaces.keys()):
            self.ctx.log(project)
        return {'projects': workspaces.keys()}

    @error_handler    
    def create(self, name):
        name = self._get_name(name)
        region = self.ctx.config.get('cluster/region', 'eastus2')
        resource_group = self.ctx.config.get(
            'resource_group', name+'-resources')
        self.ctx.log('Creating %s' % name)

        self.ws = Workspace.create(
            name=name,
            subscription_id=self.credentials.subscription_id,
            resource_group=resource_group,
            create_resource_group=True,
            location=region,
            auth=self.credentials.get_serviceprincipal_auth())
        self._select(name)
        self.ctx.log('%s created' % name)
        return {'created': name}

    @error_handler    
    def delete(self, name):
        name = self._get_name(name)
        ws = Workspace.get(
            name, 
            subscription_id=self.credentials.subscription_id, 
            auth=self.credentials.get_serviceprincipal_auth())
        self.ctx.log('Deleting %s' % name)
        ws.delete(delete_dependent_resources=True, no_wait=False)
        self._select(None)
        self.ctx.log('%s deleted' % name)
        return {'deleted': name}

    @error_handler    
    def select(self, name = None):
        self._select(name)
        self.ctx.log('Selected project %s' % name)
        return {'selected': name}

    def _select(self, name):
        self.ctx.config.set('name', name)
        self.ctx.config.write()

    def _get_name(self, name = None):
        if name is None:
            name =  self.ctx.config.get('name', None)
        if name is None:
            raise AzureException('Please provide project name...')
        return name

    def _get_ws(self, name = None, create_if_not_exist = False):
        name = self._get_name(name)
        try:
            self.ws = Workspace.get(
                name, 
                subscription_id=self.credentials.subscription_id, 
                auth=self.credentials.get_serviceprincipal_auth())
        except Exception as e:
            if create_if_not_exist:
                self.create(name)
            else:
                raise e
        return self.ws



from .base import AugerBaseApi
from .prediction import AugerPredictionApi
from .actual import AugerActualApi
from ..exceptions import AugerException


class AugerPipelineApi(AugerBaseApi):
    """Auger Pipeline API."""

    def __init__(self, ctx, experiment_api, pipeline_id=None):
        super(AugerPipelineApi, self).__init__(
            ctx, experiment_api, None, pipeline_id)

    def create(self, trial_id):
        return self._call_create({'trial_id': trial_id},
            ['creating_files', 'packaging', 'deploying'])

    def predict(self, records, features, threshold=None):
        if self.object_id is None:
            raise AugerException('Please provide Auger Pipeline id')

        status = self.properties().get('status')
        if status != 'ready':
            if 'error' in status:
                # there are following options currently:
                # [created_files_with_error, packaged_with_error,
                # deployed_with_error undeployed_with_error]
                error_message = self.properties().get('error_message')
                raise AugerException(
                    """Pipeline %s deployment has failed with following """
                    """error on Auger Cloud: `%s`""" % (
                        self.object_id, error_message))
            else:
                raise AugerException(
                    "Pipeline %s is not ready..." % self.object_id)

        prediction_api = AugerPredictionApi(self.ctx, self)
        prediction_properties = \
            prediction_api.create(records, features, threshold)
        return prediction_properties.get('result')

    def actual(self, records):
        if self.object_id is None:
            raise AugerException('Please provide Auger Pipeline id')

        actual_api = AugerActualApi(self.ctx, self)
        actual_api.create(records)



import sys

from .impl.cloud.auth import AugerAuthApi
from .credentials import Credentials
from .impl.exceptions import AugerException


class AugerAuth(object):

    def __init__(self, ctx):
        self.ctx = ctx

    def login(self, username, password, organization, url=None):
        self.credentials = Credentials(self.ctx).load(verify=False)

        try:
            self.credentials.token = None
            self.credentials.save()

            if url is None:
                url = self.credentials.api_url

            token = AugerAuthApi(self.ctx).login(
                username, password, organization, url)

            self.credentials.token = token
            self.credentials.username = username
            self.credentials.api_url = url
            self.credentials.organization = organization
            self.credentials.save()

            self.ctx.log(
                'You are now logged in on %s as %s.' % (url, username))

        except Exception as exc:
            exc_text = str(exc)
            if 'Email or password incorrect' in exc_text:
                exc_text = 'Email or password incorrect...'
            self.ctx.log(exc_text)

    def logout(self):
        self.credentials = Credentials(self.ctx).load(verify=False)

        if self.credentials.token is None:
            self.ctx.log('You are not logged in Auger.')
        else:
            self.credentials.token = None
            self.credentials.api_url = None
            self.credentials.organization = None
            self.credentials.save()
            self.ctx.log('You are logged out of Auger.')

    def whoami(self):
        self.credentials = Credentials(self.ctx).load(verify=False)

        if self.credentials.token is None:
            self.ctx.log('Please login to Auger...')
        else:
            self.ctx.log(
                '%s %s %s' % (
                    self.credentials.username,
                    self.credentials.organization,
                    self.credentials.api_url))



import json

def convert_simple_numpy_type(obj):
    import numpy as np

    if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
        np.int16, np.int32, np.int64, np.uint8,
        np.uint16, np.uint32, np.uint64)):
        return int(obj)
    elif isinstance(obj, (np.float_, np.float16, np.float32,
        np.float64)):
        return float(obj)

    # elif isinstance(obj,(np.ndarray,)): #### This is the fix
    #     return obj.tolist()

    return None

class NumpyJSONEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        res = convert_simple_numpy_type(obj)
        if res is not None:
            return res

        return json.JSONEncoder.default(self, obj)

def json_dumps_np(data, allow_nan=True):
    return json.dumps(data, cls=NumpyJSONEncoder, allow_nan=allow_nan)



import os
import json
from azureml.core.authentication import ServicePrincipalAuthentication

from a2ml.api.utils.base_credentials import BaseCredentials
from .exceptions import AzureException

class Credentials(BaseCredentials):
    """Manage credentials on user computer."""
    def __init__(self, ctx):
        super(Credentials, self).__init__(ctx, "azure")
        self.subscription_id = None
        self.directory_tenant_id = None
        self.application_client_id = None
        self.client_secret = None

    def load(self):
        content = {}

        if hasattr(self.ctx, 'credentials'):
            content = self.ctx.credentials
        elif 'AZURE_CREDENTIALS' in os.environ:
            content = os.environ.get('AZURE_CREDENTIALS', None)
            content = json.loads(content) if content else {}
        else:
            if self._credentials_file_exist():
                with open(self.creds_file, 'r') as file:
                    content = json.loads(file.read())
            else:
                azure_creds_file = os.path.abspath('%s/.azureml/auth/azureProfile.json' % os.environ.get('HOME', ''))
                if os.path.exists(azure_creds_file):
                    from a2ml.api.utils import fsclient
                    try:
                        with fsclient.open_file(azure_creds_file, "r", encoding='utf-8-sig', num_tries=0) as file:
                            res = json.load(file)
                            content = {
                                'subscription_id': res['subscriptions'][0]['id']
                            }
                    except Exception as e:
                        if self.ctx.debug:
                            import traceback
                            traceback.print_exc()

        self.subscription_id = content.get('subscription_id')
        self.directory_tenant_id = content.get('directory_tenant_id')
        self.application_client_id = content.get('application_client_id')
        self.client_secret = content.get('client_secret')

        return self

    def serialize(self):
        return {
            'subscription_id' : self.subscription_id,
            'directory_tenant_id' : self.directory_tenant_id,
            'application_client_id': self.application_client_id,
            'client_secret': self.client_secret
        }

    def get_serviceprincipal_auth(self):        
        svc_pr = None
        if self.client_secret:
            svc_pr = ServicePrincipalAuthentication(
                tenant_id=self.directory_tenant_id,
                service_principal_id=self.application_client_id,
                service_principal_password=self.client_secret
            )

        return svc_pr

    def save(self):
        self._ensure_credentials_file()

        content = self.serialize()
        with open(self.creds_file, 'w') as file:
            file.write(json.dumps(content))

    def verify(self):
        if self.subscription_id is None:
            raise AzureException('Please provide Azure subscription id...')

        return True



import os
import json

from a2ml.api.utils.base_credentials import BaseCredentials

class Credentials(BaseCredentials):
    """Manage credentials on user computer."""
    def __init__(self, ctx):
        super(Credentials, self).__init__(ctx, "auger")
        self.organization = None
        self.username = None
        self.api_url = None
        self.token = None

    def load(self, verify=True):
        if hasattr(self.ctx, 'credentials'):
            content = self.ctx.credentials
        elif 'AUGER_CREDENTIALS' in os.environ:
            content = os.environ.get('AUGER_CREDENTIALS', None)
            content = json.loads(content) if content else {}
        else:
            self._ensure_credentials_file()
            with open(self.creds_file, 'r') as file:
                content = json.loads(file.read())

        self.username = content.get('username')
        self.organization = content.get('organization')
        self.api_url = content.get('api_url', 'https://app.auger.ai')
        self.token = content.get('token')

        if verify:
            self.verify()
            
        return self

    def serialize(self):
        return {
            'organization' : self.organization,
            'api_url': self.api_url,
            'token': self.token,
            'username': self.username
        }

    def save(self):
        self._ensure_credentials_file()

        content = self.serialize()
        with open(self.creds_file, 'w') as file:
            file.write(json.dumps(content))

    def verify(self):
        if self.token is None:
            raise Exception(
                'Please provide your credentials to Auger...')
        return True



from .impl.cloud.rest_api import RestApi
from .impl.decorators import error_handler, authenticated, with_project
from .impl.model import Model
from .credentials import Credentials

class AugerModel(object):

    def __init__(self, ctx):
        self.ctx = ctx
        credentials = Credentials(ctx).load()
        self.ctx.rest_api = RestApi(
            credentials.api_url, credentials.token)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def deploy(self, project, model_id, locally):
        model_id = Model(self.ctx, project).deploy(model_id, locally)
        return {'model_id': model_id}

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def predict(self, project, filename, model_id, threshold, locally, data, columns):
        predicted = Model(self.ctx, project).predict(
            filename, model_id, threshold, locally, data, columns)
        return {'predicted': predicted}

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def actual(self, project, filename, model_id):
        Model(self.ctx, project).actual(
            filename, model_id)
        return ''



import os
import json

from .exceptions import AzureException
from .decorators import error_handler
from a2ml.api.utils.dataframe import DataFrame
from a2ml.api.utils import fsclient

class AzureModel(object):

    def __init__(self, ctx):
        super(AzureModel, self).__init__()
        self.ctx = ctx

    @error_handler
    def deploy(self, model_id, locally):
        if locally:
            model_path = self._deploy_locally(model_id)
            # self.ctx.log('Local deployment step is not required for Azure..')
            return {'model_id': model_id}
        else:
            return self._deploy_remotly(model_id)

    def _deploy_remotly(self, model_id):
        from azureml.core.model import Model
        from azureml.core.model import InferenceConfig
        from azureml.core.webservice import Webservice
        from azureml.core.webservice import AciWebservice
        from azureml.exceptions import WebserviceException
        from azureml.train.automl.run import AutoMLRun

        ws, experiment = self._get_experiment()
        iteration, run_id = self._get_iteration(model_id)

        experiment_run = AutoMLRun(experiment = experiment, run_id = run_id)
        model_run = AutoMLRun(experiment = experiment, run_id = model_id)
        model_name = model_run.properties['model_name']
        self.ctx.log('Registering model: %s' % model_id)

        description = '%s-%s' % (model_name, iteration)
        model = experiment_run.register_model(
            model_name = model_name, iteration=iteration,
            description = description, tags = None)

        script_file_name = '.azureml/score_script.py'
        model_run.download_file(
            'outputs/scoring_file_v_1_0_0.py', script_file_name)

        # Deploying ACI Service
        aci_service_name = self._aci_service_name(model_name)
        self.ctx.log('Deploying AciWebservice %s ...' % aci_service_name)

        inference_config = InferenceConfig(
            environment = model_run.get_environment(),
            entry_script = script_file_name)

        aciconfig = AciWebservice.deploy_configuration(
            cpu_cores = 1,
            memory_gb = 2,
            tags = {'type': "inference-%s" % aci_service_name},
            description = "inference-%s" % aci_service_name)

        # Remove any existing service under the same name.
        try:
            Webservice(ws, aci_service_name).delete()
            self.ctx.log('Remove any existing service under the same name...')
        except WebserviceException:
            pass

        aci_service = Model.deploy(
            ws, aci_service_name, [model], inference_config, aciconfig)
        aci_service.wait_for_deployment(True)
        self.ctx.log('%s state %s' % (aci_service_name, str(aci_service.state)))

        return {'model_id': model_id, 'aci_service_name': aci_service_name}

    @error_handler
    def predict(self, filename, model_id, threshold, locally, data, columns):
        target = self.ctx.config.get('target', None)
        predict_data = DataFrame.load(filename, target, features=columns, data=data)

        exclude_columns = self.ctx.config.get_list('exclude')
        if exclude_columns:
            predict_data.drop(columns=exclude_columns, inplace=True, errors='ignore')

        y_pred = []
        if locally:
            y_pred, y_proba, proba_classes = self._predict_locally(
                predict_data, model_id, threshold)
        else:
            y_pred, y_proba, proba_classes = self._predict_remotely(
                predict_data, model_id, threshold)

        predict_data[target] = y_pred

        if y_proba is not None:
            for idx, name in enumerate(proba_classes):
                predict_data['proba_'+str(name)] = list(y_proba[:,idx])

        if filename:        
            predicted = self._save_predictions(predict_data, filename)
        elif columns:
            predicted = predict_data.to_dict('split')
            predicted = {'data': predicted.get('data', []), 'columns': predicted.get('columns')}
        else:
            predicted = predict_data.to_dict('records')

        return {'predicted': predicted}

    @error_handler        
    def actual(self, filename, model_id):
        pass

    def _get_iteration(self, model_id):
        iteration = None
        run_id = model_id
        parts = model_id.split('_')
        if len(parts) > 2:
            run_id = parts[0]+"_"+parts[1]
            iteration = parts[2]
        return iteration, run_id

    def _aci_service_name(self, model_name):
        # It must only consist of lowercase letters, numbers, or dashes, start
        # with a letter, end with a letter or number, and be between 3 and 32
        # characters long.
        #TODO - service_name + suffix must satisfy requiremets
        if model_name.endswith('-service'):
            return model_name
            
        return (model_name+'-service').lower()

    def _get_experiment(self):
        from azureml.core import Experiment
        from .project import AzureProject

        ws = AzureProject(self.ctx)._get_ws()
        experiment_name = self.ctx.config.get('experiment/name', None)
        if experiment_name is None:
            raise AzureException('Please specify Experiment name...')
        experiment = Experiment(ws, experiment_name)

        return ws, experiment

    def _predict_remotely(self, predict_data, model_id, threshold):
        from azureml.core.webservice import AciWebservice

        ws, experiment = self._get_experiment()

        input_payload = predict_data.to_json(orient='split', index = False)

        if model_id.startswith("AutoML_"):
            from azureml.train.automl.run import AutoMLRun

            remote_run = AutoMLRun(experiment = experiment, run_id = model_id)
            model_name = remote_run.properties['model_name']
        else:
            model_name = model_id

        aci_service_name = self._aci_service_name(model_name)
        aci_service = AciWebservice(ws, aci_service_name)

        input_payload = json.loads(input_payload)
        # If you have a classification model, you can get probabilities by changing this to 'predict_proba'.        
        method = 'predict'
        if threshold is not None:
            method = 'predict_proba'
        input_payload = {
            'method': method,
            'data': input_payload['data']
        }
        input_payload = json.dumps(input_payload)
        try:
            response = aci_service.run(input_data = input_payload)
        except Exception as e:
            # print('err log', aci_service.get_logs())
            raise e

        results_proba = None
        proba_classes = None

        return json.loads(response)['result'], results_proba, proba_classes

    def verify_local_model(self, model_id):
        model_path = os.path.join(self.ctx.config.get_path(), 'models',
            'azure-model-%s.pkl.gz'%model_id)
        return fsclient.is_file_exists(model_path), model_path
            
    def _deploy_locally(self, model_id):
        is_loaded, model_path = self.verify_local_model(model_id)

        if not is_loaded:
            from azureml.train.automl.run import AutoMLRun

            self.ctx.log('Downloading model %s' % model_id)

            ws, experiment = self._get_experiment()
            iteration, run_id = self._get_iteration(model_id)
            remote_run = AutoMLRun(experiment = experiment, run_id = run_id)
            best_run, fitted_model = remote_run.get_output(iteration=iteration)
            fsclient.save_object_to_file(fitted_model, model_path)

            self.ctx.log('Downloaded model to %s' % model_path)
        else:
            self.ctx.log('Downloaded model is %s' % model_path)

        return model_path
                
    def _predict_locally(self, predict_data, model_id, threshold):
        model_path = self._deploy_locally(model_id)
        fitted_model = fsclient.load_object_from_file(model_path)
        try:
            model_features = list(fitted_model.steps[0][1]._columns_types_mapping.keys())
            predict_data = predict_data[model_features]
            predict_data.to_csv("test_options.csv", index=False, compression=None, encoding='utf-8')
        except Exception as e:
            self.ctx.log('Cannot get columns from model.Use original columns from predicted data: %s'%e)

        results_proba = None
        proba_classes = None
        if threshold is not None:
            results_proba = fitted_model.predict_proba(predict_data)
            proba_classes = list(fitted_model.classes_)
            minority_target_class = self.ctx.config.get('minority_target_class', None)
            if minority_target_class is None:
                for item in proba_classes:
                    if item == 1 or item == True or item == "1" or item == "True":
                        minority_target_class = item
                            
            result = self._calculate_proba_target(results_proba,
                proba_classes, None, threshold, minority_target_class)
        else:
            result = fitted_model.predict(predict_data)

        return result, results_proba, proba_classes

    def _calculate_proba_target(self, results_proba, proba_classes, proba_classes_orig, threshold, minority_target_class=None):
        import json
        results = []

        if type(threshold) == str:
            try:
                threshold = float(threshold)
            except:
                try:
                    threshold = json.loads(threshold)
                except Exception as e:
                    raise Exception("Threshold '%s' should be float or hash with target classes. Error: %s"%(threshold, str(e)))

        if type(threshold) != dict and minority_target_class is not None:
            threshold = {minority_target_class:threshold}

        # print("Prediction threshold: %s, %s"%(threshold, proba_classes_orig))
        #print(results_proba)
        if type(threshold) == dict:
            mapped_threshold = {}
            if not proba_classes_orig:
                proba_classes_orig = proba_classes

            for name, value in threshold.items():
                idx_class = None
                for idx, item in enumerate(proba_classes_orig):
                    if item == name:
                        idx_class = idx
                        break

                if idx_class is None:
                    raise Exception("Unknown target class in threshold: %s, %s"%(name, proba_classes_orig))

                mapped_threshold[idx_class] = value

            for item in results_proba:
                proba_idx = None
                for idx, value in mapped_threshold.items():
                    if item[idx] >= value:
                        proba_idx = idx
                        break

                if proba_idx is None:
                    proba_idx = 0
                    for idx, value in enumerate(item):
                        if idx not in mapped_threshold:
                            proba_idx = idx
                            break

                results.append(proba_classes[proba_idx])
        else:
            #TODO: support multiclass classification
            for item in results_proba:
                max_proba_idx = 0
                for idx, prob in enumerate(item):
                    if prob > item[max_proba_idx]:
                        max_proba_idx = idx

                if item[max_proba_idx] < threshold:
                    if max_proba_idx > 0:
                        max_proba_idx = 0
                    else:
                        max_proba_idx = 1

                results.append(proba_classes[max_proba_idx])

        return results
             
    def _save_predictions(self, df_predictions, filename):
        predicted_path = os.path.abspath(
            os.path.splitext(filename)[0] + "_predicted.csv")
        df_predictions.to_csv(predicted_path, index=False, encoding='utf-8')
        self.ctx.log('Predictions are saved to %s' % predicted_path)
        return predicted_path



from a2ml.api.base_a2ml import BaseA2ML
from a2ml.api.utils.show_result import show_result

class A2MLModel(BaseA2ML):
    """Contains the model operations that interact with provider."""
    def __init__(self, ctx, provider):
        """Initializes a new a2ml model.

        Args:
            context (object): An instance of the a2ml Context.
            provider (str): The automl provider(s) you wish to run. For example 'auger,azure,google'.
        
        Returns:
            A2MLModel object

        Examples:
            .. code-block:: python

                ctx = Context()
                model = A2MLModel(ctx, 'auger, azure')
        """
        super(A2MLModel, self).__init__()
        self.ctx = ctx
        self.provider = provider
        self.runner = self.build_runner(ctx, provider, 'model')
        self.local_runner = lambda: self.build_runner(ctx, provider, 'model', force_local=True)

    @show_result
    def deploy(self, model_id, locally):
        """Deploy a model locally or to specified provider(s).

        Args:
            model_id (str): Model ID from the any experiment leaderboard.
            locally(bool): Deploys using a local model if True, on the Provider Cloud if False.
        
        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'data': {'model_id': 'D881079E1ED14FB'}
                    }
                }

        Status:

                preprocess: Search is preprocessing data for traing
                started: Search is in progress
                completed: Search is completed
                interrupted: Search was interrupted
                
        Examples:
            .. code-block:: python

                ctx = Context()
                model = Model(ctx, 'auger,azure').deploy(model_id='D881079E1ED14FB', locally=True)
        """
        return self.__get_runner(locally).execute('deploy', model_id, locally)

    @show_result
    def predict(self, filename, model_id, threshold=None, locally=False, data=None, columns=None):
        """Predict results with new data against deployed model. Predictions are stored next to the file with data to be predicted on. The file name will be appended with suffix _predicted.

        Args:
            filename(str): The file with data to request predictions for.
            model_id(str): The deployed model id you want to use.
            threshold(float): For classification models only. This will return class probabilities with response.
            locally(bool): Predicts using a local model if True, on the Provider Cloud if False.
        
        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'data': {'model_id': 'D881079E1ED14FB'}
                    }
                }


        Examples:
            .. code-block:: python

                ctx = Context()
                model = A2MLModel(ctx, 'auger, azure').predict(filename=<path_to_file>/dataset.csv,model_id='D881079E1ED14FB',threshold=None,locally=False)
        """
        return self.__get_runner(locally).execute('predict', filename, model_id, threshold, locally, data, columns)

    @show_result
    def actual(self, filename, model_id):
        """Submits actual results(ground truths) for predictions of a deployed model. This is used to review and monitor active models.

        Note:
            It is assumed you have predictions against this model first. The file will need to fill in actual values for prediction_id. 
            
            .. list-table:: actuals.csv
                :widths: 50 50
                :header-rows: 1

                * - prediction_id
                  - actual
                * - eaed9cd8-ba49-4c06-86d5-71d453c681d1
                  - Iris-setosa
                



        Args:
            filename(str): The file with data to request predictions for.
            model_id(str): The deployed model id you want to use.
        
        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'data': ''
                    }
                }


        Examples:
            .. code-block:: python

                ctx = Context()
                model = A2MLModel(ctx, 'auger, azure').actual(filename=<path_to_file>/dataset_actuals.csv,model_id='D881079E1ED14FB')
        """
        return self.runner.execute('actual', filename, model_id)

    def __get_runner(self, locally):
        if locally:
            return self.local_runner()
        else:
            return self.runner



import pandas

from  a2ml.api.utils import fsclient


class DataFrame(object):
    """Warpper around Pandas DataFrame."""

    def __init__(self):
        super(DataFrame, self).__init__()

    @staticmethod
    def load(filename, target, features=None, nrows=None, data=None):
        df = None

        if filename:
            if filename.endswith('.json') or filename.endswith('.json.gz'):
                df = pandas.read_json(filename)
            elif filename.endswith('.xlsx') or filename.endswith('.xls'):
                df = pandas.read_excel(filename)
            elif filename.endswith('.feather') or filename.endswith('.feather.gz'):
                import feather
                with fsclient.open_file(filename, 'rb', encoding=None) as local_file:
                    df = feather.read_dataframe(local_file, columns=features, use_threads=bool(True))

            if df is None:        
                try:
                    df = DataFrame._read_csv(filename, ',', features, nrows)
                except Exception as e:
                    df = DataFrame._read_csv(filename, '|', features, nrows)

        else:
            df = DataFrame.load_data(data, features)

        features = df.columns.tolist()
        if target in features:
            df.drop(columns=[target], inplace=True)

        return df

    @staticmethod
    def load_records(filename, target, features=None, nrows=None, data=None):
        df = DataFrame.load(filename, target, features, nrows, data)

        features = df.columns.tolist()
        df.replace({pandas.np.nan: None}, inplace=True)
        records = df.values.tolist()

        return records, features

    @staticmethod
    def load_data(data, columns):
        df = None
        if columns:
            df = pandas.DataFrame.from_records(data, columns=columns)
        else:
            df = pandas.DataFrame(data)

        return df
            
    @staticmethod
    def save(filename, data):
        df = pandas.DataFrame.from_records(data['data'], columns=data['columns'])

        fsclient.create_parent_folder(filename)
        df.to_csv(filename, index=False, encoding='utf-8')

    @staticmethod
    def convert_records_to_dict(data):
        df = pandas.DataFrame.from_records(data['data'], columns=data['columns'])
        return df.to_dict('records')

    @staticmethod
    def save_df(filename, df):
        fsclient.create_parent_folder(filename)
        
        df.to_csv(filename, index=False, encoding='utf-8')

    @staticmethod
    def _read_csv(filename, sep, features=None, nrows=None):
        return pandas.read_csv(filename,
            encoding='utf-8', escapechar="\\", usecols=features,
            na_values=['?'], header=0, prefix=None, sep=sep,
            nrows=nrows, low_memory=False, compression='infer')




class AzureA2ML(object):
    """Auger A2ML implementation."""

    def __init__(self, ctx):
        super(AzureA2ML, self).__init__()
        self.ctx = ctx

    def import_data(self):
        from a2ml.api.azure.dataset import AzureDataset

        return AzureDataset(self.ctx).create()

    def train(self):
        from a2ml.api.azure.experiment import AzureExperiment

        return AzureExperiment(self.ctx).start()

    def evaluate(self, run_id = None):
        from a2ml.api.azure.experiment import AzureExperiment

        return AzureExperiment(self.ctx).leaderboard(run_id)

    def deploy(self, model_id, locally=False):
        from a2ml.api.azure.model import AzureModel

        return AzureModel(self.ctx).deploy(model_id, locally)

    def predict(self, filename, model_id, threshold=None, locally=False, data=None, columns=None):
        from a2ml.api.azure.model import AzureModel
        
        return AzureModel(self.ctx).predict(
            filename, model_id, threshold, locally, data, columns)

    def review(self):
        pass



import re
import sys
import time

from auger.hub_api_client import HubApiClient
from ..exceptions import AugerException


REQUEST_LIMIT = 100
STATE_POLL_INTERVAL = 10

class RestApi(object):
    """Warapper around Auger Cloud Rest Api."""

    def __init__(self, url, token, debug=False):
        super(RestApi, self).__init__()
        self.hub_client = HubApiClient(hub_app_url=url, token=token, debug=debug)
        self.api_url = url
        self.debug = debug

    def get_status(self, obj, obj_id):
        return self.hub_client.get_status(object=obj, id=obj_id)

    def call_ex(self, method, params={}):
        params = params.copy()

        if params.get('id') and not method.startswith('create_'):
            oid = params['id']
            del params['id']
            return getattr(self.hub_client, method)(oid, **params)
        else:
            return getattr(self.hub_client, method)(**params)

    def call(self, method, params={}):
        result = self.call_ex(method, params)

        if 'data' in result:
            return result['data']

        raise AugerException("Call of Auger API method %s failed." % keys)

    def request_list(self, record_type, params):
        offset = params.get('offset', 0)
        limit = params.get('limit', REQUEST_LIMIT)
        p = params.copy()
        while limit > 0:
            p['offset'] = offset
            p['limit'] = limit
            response = self.call_ex('get_' + record_type, p)
            try:
                if not 'data' in response or not 'meta' in response:
                    raise AugerException("Read list of %s failed." % record_type)

                for item in response['data']:
                    yield item

                received = len(response['data'])
                offset += received
                limit -= received
                if offset >= response['meta']['pagination']['total']:
                    break
            except Exception as e:
                    import traceback; traceback.print_exc()

    def wait_for_object_status(self,
        get_status, progress, object_readable_name,
        post_check_status=None, log_status=None):

        def _log_status(obj_status): pass
        log_status  = log_status if log_status else _log_status

        status = get_status()
        last_status = ''

        while status in progress:
            if status != last_status:
                last_status = status
                log_status(status)

            while status == last_status:
                time.sleep(STATE_POLL_INTERVAL)
                status = get_status()

        if status == 'processed_with_error':
            raise AugerException(
                '%s processed with error' % object_readable_name)
        elif status == 'error' or status == "failure":
            raise AugerException('Auger Cloud return error...')

        if post_check_status:
            post_check_status(status)

        return status



from a2ml.api.utils.crud_runner import CRUDRunner
from a2ml.api.utils.provider_runner import ProviderRunner
from a2ml.api.utils.remote_runner import RemoteRunner


class BaseA2ML(object):
    def build_runner(self, ctx, provider, object_name=None, force_local=False):
        if ctx.config.get('use_server') == True and not force_local:
            return RemoteRunner(ctx, provider, object_name)
        elif object_name:
            return CRUDRunner(ctx, provider, object_name)
        else:
            return ProviderRunner(ctx, provider)



import click
from a2ml.api.a2ml import A2ML
from a2ml.api.utils.context import pass_context


@click.command('evaluate', short_help='Evaluate models after training.')
@click.argument('run-id', required=False, type=click.STRING)
@click.option('--provider', '-p', type=click.STRING, required=False,
    help='Cloud AutoML Provider.')
@pass_context
def cmdl(ctx, run_id, provider):
    """Evaluate models after training."""
    ctx.setup_logger(format='')
    A2ML(ctx, provider).evaluate(run_id)



from .impl.project import Project
from .impl.exceptions import AugerException
from .impl.decorators import error_handler, authenticated
from .impl.cloud.rest_api import RestApi
from .credentials import Credentials
from .config import AugerConfig

class AugerProject(object):

    def __init__(self, ctx):
        self.ctx = ctx
        credentials = Credentials(ctx).load()
        self.ctx.rest_api = RestApi(
            credentials.api_url, credentials.token)

    @error_handler
    @authenticated
    def list(self):
        count = 0
        for project in iter(Project(self.ctx).list()):
            self.ctx.log(project.get('name'))
            count += 1
        self.ctx.log('%s Project(s) listed' % str(count))
        return {'projects': Project(self.ctx).list()}

    @error_handler
    @authenticated
    def create(self, name):
        old_name, name, project = self._setup_op(name, False)
        project.create()
        if name != old_name:
            self._set_project_config(name)
        self.ctx.log('Created Project %s' % name)
        return {'created': name}

    @error_handler
    @authenticated
    def delete(self, name):
        old_name, name, project = self._setup_op(name)
        project.delete()
        if name == old_name:
            self._set_project_config(None)
        self.ctx.log('Deleted Project %s' % name)
        return {'deleted': name}

    @error_handler
    @authenticated
    def start(self, name):
        old_name, name, project = self._setup_op(name)
        if not project.is_running():
            self.ctx.log('Starting Project...')
            project.start()
            self.ctx.log('Started Project %s' % name)
        else:
            self.ctx.log('Project is already running...')
        return {'running': name}

    @error_handler
    @authenticated
    def stop(self, name):
        old_name, name, project = self._setup_op(name)
        if project.is_running():
            self.ctx.log('Stopping Project...')
            project.stop()
            self.ctx.log('Stopped Project %s' % name)
        else:
            self.ctx.log('Project is not running...')
        return {'stopped': name}

    @error_handler
    @authenticated
    def select(self, name):
        old_name, name, project = self._setup_op(name, False)
        if name != old_name:
            self._set_project_config(name)
        self.ctx.log('Selected Project %s' % name)
        return {'selected': name}

    def _set_project_config(self, name):
        AugerConfig(self.ctx).\
            set_project(name).\
            set_data_set(None).\
            set_experiment(None, None)

    def _setup_op(self, name, verify_project=True):
        old_name = self.ctx.config.get('name', None)
        if name is None:
            name = old_name
        if name is None:
            raise AugerException('Please specify Project name...')

        project = Project(self.ctx, name)

        if verify_project and not project.is_exists:
            raise AugerException('Project %s doesn\'t exists...' % name)

        return old_name, name, project



import os
import sys
import errno
import click

from a2ml.api.utils.context import PROVIDERS
from a2ml.cmdl.utils.template import Template
from a2ml.api.utils.context import pass_context
from a2ml.api.auger.impl.cloud.dataset import AugerDataSetApi
from a2ml.api.utils import fsclient


class NewCmd(object):

    def __init__(self, ctx, project_name,
        providers, target, source, model_type):
        self.ctx = ctx
        self.project_name = project_name
        self.providers = providers
        self.target = target
        self.source = source
        self.model_type = model_type

    def mk_project_folder(self):
        if not self.ctx.config.path:
            project_path = os.path.join(os.getcwd(), self.project_name)
        else:
            project_path = self.ctx.config.path

        if not fsclient.is_s3_path(project_path):
            project_path = os.path.abspath(project_path)

        if fsclient.is_folder_exists(project_path):
            raise Exception(
                'Can\'t create \'%s\'. Folder already exists.' % \
                    self.project_name)

        fsclient.create_folder(project_path)    
        self.ctx.log('Created project folder %s', self.project_name)
        return project_path

    def create_project(self):
        try:
            if self.source:
                self.source = AugerDataSetApi.verify(self.source)[0]

            project_path = self.mk_project_folder()
            Template.copy_config_files(project_path, ['config'] + PROVIDERS)
            self.ctx.log('To build your model, please do:'
                ' cd %s && a2ml import && a2ml train' % \
                self.project_name)

            config = self.ctx.config
            config.load(project_path, True)
            config.set('providers',
                PROVIDERS if self.providers == 'all' else self.providers, config_name='config')
            config.set('name', self.project_name, config_name='config')
            config.set('target', self.target, config_name='config')
            config.set('source', self.source, config_name='config')
            config.set('model_type', self.model_type, config_name='config')
            config.write('config')

            if self.model_type != 'classification':
                self.ctx.config.set('experiment/metric', 'spearman_correlation', config_name='azure')
                self.ctx.config.write('azure')
                self.ctx.config.set('experiment/metric', 'r2', config_name='auger')
                self.ctx.config.write('auger')

        except Exception as e:
            import traceback
            traceback.print_exc()
            self.ctx.log('%s', str(e))
            #sys.exit(1)


@click.command('new', short_help='Create new A2ML project.')
@click.argument('project', required=True, type=click.STRING)
@click.option('--providers', '-p', default='auger',
    type=click.Choice(['all','auger','google','azure']),
    help='Model Provider.')
@click.option('--source', '-s',  default='', type=click.STRING,
    help='Data source local file or remote url.')
@click.option('--model-type', '-mt', default='classification',
    type=click.Choice(['classification','regression','timeseries']),
    help='Model Type.')
@click.option('--target', '-t',  default='', type=click.STRING,
    help='Target column name in data source.')
@pass_context
def cmdl(ctx, project, providers, source, model_type, target):
    """Create new A2ML project."""
    ctx.setup_logger(format='')
    NewCmd(ctx, project, providers,
        target, source, model_type).create_project()



from .rest_api import RestApi
from .org import AugerOrganizationApi
from ..exceptions import AugerException


class AugerAuthApi(object):
    """Auger Authentication API."""
    def __init__(self, ctx):
        super(AugerAuthApi, self).__init__()
        self.ctx = ctx

    def login(self, username, password, organization, url):
        rest_api = RestApi(url, None)
        try:
            res = rest_api.call_ex(
                'create_token', {'email': username, 'password': password})
        except Exception as e:
            if 'Email or password incorrect' in str(e):
                raise AugerException('Email or password incorrect...')
            else:
                raise e
        self.ctx.rest_api = RestApi(url, res['data']['token'])
        org_api = AugerOrganizationApi(self.ctx, organization)
        if org_api.properties() == None:
            raise AugerException(
                'Auger Organization %s doesn\'t exist' % organization)
        return res['data']['token']



import os
import codecs
import sys

from setuptools import setup
from setuptools import find_packages
from setuptools.command.install import install

VERSION = '0.3.8'

# Get the long description from the README file
here = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


class VerifyVersionCommand(install):
    """Verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG', '')

        if not tag.endswith(VERSION, 1):
            info = "Git tag: {0} doesn't match version of a2ml: {1}".format(
                tag, VERSION
            )
            sys.exit(info)

install_requires = [
    'numpy<=1.16.2,>=1.16.0', #version for azure
    'pandas<=0.23.4,>=0.21.0', #version for azure
    'joblib==0.14.1', #version for azure
    'ruamel.yaml>0.16.7', #version for azure
    'auger-hub-api-client>=0.6.1',
    'click',
    'shortuuid',
    'docutils<0.16,>=0.10',
    'requests',
    'smart_open==1.9.0',
    'jsonpickle',
    'feather-format'
]

extras = {
    'testing': [
        'flake8',
        'mock',
        'pytest',
        'pytest-cov',
        'pytest-runner',
        'pytest-xdist',
        'tox',
        'twine',
        'wheel>=0.30.0,<0.31.0'
    ],
    'docs': [
        'sphinx'
    ],
    'server': [
        'celery==4.4.0',
        'aioredis',
        'asyncio',
        'fastapi',
        'gevent',
        'redis',
        'uvicorn',
        's3fs',
        'boto3'
    ],
    'azure': [
        'azureml-sdk[automl]~=1.5.0'
    ],
    'azure-predict-local': [
        'sklearn-pandas<=1.7.0,>=1.4.0',
        'azureml-automl-runtime~=1.5.0'
    ],
    'azure-predict-remote': [
        'azureml-core~=1.5.0'
    ],
    'azure-deploy-predict': [
        'sklearn-pandas<=1.7.0,>=1.4.0',
        'azureml-automl-runtime~=1.5.0',
        'azureml-train-automl-client~=1.5.0',
        'azureml-train-automl-runtime~=1.5.0'
    ],
    'google': [
        'google-cloud-automl'
    ]
}

# Meta dependency groups.
all_deps = []
for group_name in extras:
    if group_name != 'azure-predict-local' and group_name != 'azure-predict-remote' and group_name != 'azure-deploy-predict':
        all_deps += extras[group_name]
extras['all'] = all_deps

description = """A2ML ("Automate AutoML") is a set of scripts to automate
 Automated Machine Learning workflows from multiple vendors."""

setup(
    name='a2ml',
    version=VERSION,
    description=description,
    long_description=description,
    # TODO: Convert readme to rst for azureml / wheels bug
    # long_description=long_description,
    # long_description_content_type='text/markdown',
    author='Auger AI',
    author_email='hello@auger.ai',
    url='https://github.com/augerai/a2ml',
    zip_safe=False,
    platforms='any',
    test_suite='tests',
    python_requires='>=3',
    keywords='augerai aa2ml.cmdl.cmdl:cmdluger ai '
        'machine learning automl deeplearn api sdk',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        "Intended Audience :: System Administrators",
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        "Programming Language :: Python :: 3 :: Only"
    ],
    install_requires=install_requires,
    extras_require=extras,
    tests_require=extras['testing'],
    entry_points={
        'console_scripts': [
            'a2ml=a2ml.cmdl.cmdl:cmdl'
        ]
    },
    cmdclass={
        'verify': VerifyVersionCommand
    },
    packages=find_packages(),
    package_data={
        'a2ml': [
            'api/azure/*.template',
            'cmdl/template/*.template',
            'cmdl/template/*.yaml'
        ]
    }
)



import os
import pytest
import shutil
import logging
from click.testing import CliRunner

@pytest.fixture
def runner():
    return CliRunner()

@pytest.fixture(scope="function")
def isolated(runner):
    with runner.isolated_filesystem():
        yield runner

@pytest.fixture
def log(caplog):
    caplog.set_level(logging.INFO)
    return caplog

@pytest.fixture
def project(isolated):
    source = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), '..', 'fixtures',
        'cli-integration-test')
    shutil.copytree(source, './cli-integration-test')
    os.chdir('cli-integration-test')



from .base import AugerBaseApi
from ..exceptions import AugerException


class AugerPredictionApi(AugerBaseApi):
    """Auger Trial API."""

    def __init__(self, ctx, pipeline_api,
        prediction_name=None, prediction_id=None):
        super(AugerPredictionApi, self).__init__(
            ctx, pipeline_api, prediction_name, prediction_id)
        assert pipeline_api is not None, 'Pipeline must be set for Prediction'

    def create(self, records, features, threshold=None):
        return self._call_create({
            'pipeline_id': self.parent_api.object_id,
            'records': records, 
            'features': features,
            'threshold': threshold
            }, ['requested', 'running'])



from fastapi import FastAPI, Request, WebSocket
from fastapi.responses import HTMLResponse, PlainTextResponse, JSONResponse
from fastapi.encoders import jsonable_encoder

import asyncio
import uuid
import websockets
import sys

from a2ml.tasks_queue.tasks_api import new_project_task, list_projects_task, delete_project_task, select_project_task, \
    new_dataset_task, list_datasets_task, delete_dataset_task, select_dataset_task, \
    list_experiments_task, leaderboard_experiment_task, history_experiment_task, \
    start_experiment_task, stop_experiment_task, \
    actual_model_task, deploy_model_task, predict_model_task, \
    deploy_task, evaluate_task, import_data_task, predict_task, train_task, review_task

from a2ml.server.notification import AsyncReceiver

app = FastAPI()

API_SCHEMA = {
    '/api/v1/datasets': {
        'get': list_datasets_task,
        'post': new_dataset_task,
        'delete': delete_dataset_task,
    },
    '/api/v1/datasets/select': {
        'patch': select_dataset_task,
    },
    '/api/v1/deploy': {
        'patch': deploy_task,
    },
    '/api/v1/evaluate': {
        'patch': evaluate_task,
    },
    '/api/v1/experiments': {
        'get': list_experiments_task,
    },
    '/api/v1/experiments/history': {
        'get': history_experiment_task,
    },
    '/api/v1/experiments/leaderboard': {
        'get': leaderboard_experiment_task,
    },
    '/api/v1/experiments/start': {
        'patch': start_experiment_task,
    },
    '/api/v1/experiments/stop': {
        'patch': stop_experiment_task,
    },
    '/api/v1/import_data': {
        'patch': import_data_task,
    },
    '/api/v1/models/actual': {
        'post': actual_model_task,
    },
    '/api/v1/models/deploy': {
        'patch': deploy_model_task,
    },
    '/api/v1/models/predict': {
        'post': predict_model_task,
    },
    '/api/v1/predict': {
        'patch': predict_task,
    },
    '/api/v1/projects': {
        'get': list_projects_task,
        'post': new_project_task,
        'delete': delete_project_task,
    },
    '/api/v1/projects/select': {
        'patch': select_project_task,
    },
    '/api/v1/review': {
        'patch': review_task,
    },
    '/api/v1/train': {
        'patch': train_task,
    },
}

@app.get('/hello')
async def say_hello():
    await asyncio.sleep(1)
    return {'Hello': 'World'}

def define_endpoints(schema):
    for path in schema.keys():
        endpoints = schema[path]
        for http_verb in endpoints.keys():
            task = endpoints[http_verb]
            define_endpoint(http_verb, path, task)

def define_endpoint(http_verb, path, task):
    decorator_method = getattr(app, http_verb)

    @decorator_method(path)
    async def handler(request: Request):
        return await __run_task(task, request)

    setattr(sys.modules[__name__], task.__name__.replace('_task', ''), handler)

define_endpoints(API_SCHEMA)

def log(*args):
    print(*args)

async def __run_task(task, request):
    request_id = __generate_request_id()
    params = await __get_body_and_query_params(request)
    params['_request_id'] = request_id
    task.delay(params)
    return __render_response(request_id)

async def __get_body_and_query_params(request):
    params = dict(request.query_params)
    if len(await request.body()) > 0:
        params.update(await request.json())

    return params

def __render_response(request_id):
    res = {
        'meta': { 'status': 200 },
        'data': { 'request_id': request_id },
    }

    content = jsonable_encoder(res)
    return JSONResponse(content=content)

def __generate_request_id():
    return str(uuid.uuid4())

@app.websocket("/ws")
# id - id of request
async def websocket_endpoint(websocket: WebSocket, id: str = None, last_msg_id: str = '0'):
    try:
        if id:
            await websocket.accept()
            await websocket.send_json({"type": "start", 'request_id': id}, mode="text")

            try:
                notificator = AsyncReceiver(id, last_msg_id)
                try:
                    while True:
                        try:
                            # Periodically iterrupt waiting of message from subscription
                            # to check is websocket is still alive or not
                            reply = await asyncio.wait_for(notificator.get_message(), timeout=5.0)
                            if reply:
                                log("Broadcast: ", repr(reply))
                                await websocket.send_text(reply)
                            else:
                                await websocket.send_json({"type": "ping"}, mode="text")
                        except asyncio.TimeoutError:
                            await websocket.send_json({"type": "ping"}, mode="text")
                except (websockets.exceptions.ConnectionClosedOK, websockets.exceptions.ConnectionClosedError) as e:
                    log(f"WebSocket {id} disconnected: {str(e)}")
            finally:
                await notificator.close()
    finally:
        log('WebSocket stopped')
        await websocket.close()



from .celery_app import celeryApp
import logging
import copy
import os
import json
import jsonpickle

from a2ml.api.utils.context import Context
from a2ml.api.a2ml import A2ML
from a2ml.api.a2ml_dataset import A2MLDataset
from a2ml.api.a2ml_experiment import A2MLExperiment
from a2ml.api.a2ml_model import A2MLModel
from a2ml.api.a2ml_project import A2MLProject
from a2ml.server.notification import SyncSender

notificator = SyncSender()

def create_context(params, new_project=False):
    if params.get('context'):
        ctx = jsonpickle.decode(params['context'])

        ctx.set_runs_on_server(True)
        ctx.config.set('use_server', False, config_name='config')
        ctx.notificator = notificator
        ctx.request_id = params['_request_id']
        ctx.setup_logger(format='')
    else:
        ctx = Context(
            path=params.get('project_path'),
            debug=params.get('debug_log', False)
        )

        if not new_project:
            if params.get('provider'):
                ctx.config.set('providers', [params.get('provider')], config_name='config')

            if params.get('source_path'):
                ctx.config.set('source', params.get('source_path'), config_name='config')

    return ctx

def __handle_task_result(self, status, retval, task_id, args, kwargs, einfo):
    request_id = args[0]['_request_id']

    if status == 'SUCCESS':
        notificator.publish_result(request_id, status, retval)
    else:
        notificator.publish_result(
            request_id,
            status,
            __error_to_result(retval, einfo)
        )

# Projects
@celeryApp.task(after_return=__handle_task_result)
def new_project_task(params):
    return with_context(
        params,
        lambda ctx: A2MLProject(ctx, None).create(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def list_projects_task(params):
    def func(ctx):
        res = A2MLProject(ctx, None).list(*params['args'], **params['kwargs'])
        return __map_collection_to_name(res, 'projects')

    return with_context(params, func)

@celeryApp.task(after_return=__handle_task_result)
def delete_project_task(params):
    return with_context(
        params,
        lambda ctx: A2MLProject(ctx, None).delete(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def select_project_task(params):
    return with_context(
        params,
        lambda ctx: A2MLProject(ctx, None).select(*params['args'], **params['kwargs'])
    )

# Datasets
@celeryApp.task(after_return=__handle_task_result)
def new_dataset_task(params):
    return with_context(
        params,
        lambda ctx: A2MLDataset(ctx, None).create(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def list_datasets_task(params):
    def func(ctx):
        res = A2MLDataset(ctx, None).list(*params['args'], **params['kwargs'])
        return __map_collection_to_name(res, 'datasets')

    return with_context(params, func)

@celeryApp.task(after_return=__handle_task_result)
def delete_dataset_task(params):
    return with_context(
        params,
        lambda ctx: A2MLDataset(ctx, None).delete(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def select_dataset_task(params):
    return with_context(
        params,
        lambda ctx: A2MLDataset(ctx, None).select(*params['args'], **params['kwargs'])
    )

# Experiment
@celeryApp.task(after_return=__handle_task_result)
def list_experiments_task(params):
    def func(ctx):
        res = A2MLExperiment(ctx, None).list(*params['args'], **params['kwargs'])
        return __map_collection_to_name(res, 'experiments')

    return with_context(params, func)

@celeryApp.task(after_return=__handle_task_result)
def leaderboard_experiment_task(params):
    return with_context(
        params,
        lambda ctx: A2MLExperiment(ctx, None).leaderboard(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def history_experiment_task(params):
    return with_context(
        params,
        lambda ctx: A2MLExperiment(ctx, None).history(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def start_experiment_task(params):
    return with_context(
        params,
        lambda ctx: A2MLExperiment(ctx, None).start(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def stop_experiment_task(params):
    return with_context(
        params,
        lambda ctx: A2MLExperiment(ctx, None).stop(*params['args'], **params['kwargs'])
    )

# Models
@celeryApp.task(after_return=__handle_task_result)
def actual_model_task(params):
    return with_context(
        params,
        lambda ctx: A2MLModel(ctx, None).actual(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def deploy_model_task(params):
    return with_context(
        params,
        lambda ctx: A2MLModel(ctx, None).deploy(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def predict_model_task(params):
    return with_context(
        params,
        lambda ctx: A2MLModel(ctx, None).predict(*params['args'], **params['kwargs'])
    )

# Complex tasks
@celeryApp.task(after_return=__handle_task_result)
def import_data_task(params):
    return with_context(
        params,
        lambda ctx: A2ML(ctx).import_data(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def train_task(params):
    return with_context(
        params,
        lambda ctx: A2ML(ctx).train(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def evaluate_task(params):
    return with_context(
        params,
        lambda ctx: A2ML(ctx).evaluate(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def deploy_task(params):
    return with_context(
        params,
        lambda ctx: A2ML(ctx).deploy(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def predict_task(params):
    return with_context(
        params,
        lambda ctx: A2ML(ctx).predict(*params['args'], **params['kwargs'])
    )

@celeryApp.task(after_return=__handle_task_result)
def review_task(params):
    # TODO
    raise Exception('not inplemented yet')

@celeryApp.task(after_return=__handle_task_result)
def demo_task(params):
    import time
    request_id = params['_request_id']
    for i in range(0, 10):
        notificator.publish_log(request_id, 'info', 'log ' + str(i))
        time.sleep(2)

    notificator.publish_result(request_id, 'SUCCESS', 'done')

def with_context(params, proc):
    ctx = create_context(params)

    if 'args' not in params:
        params['args'] = []

    if 'kwargs' not in params:
        params['kwargs'] = {}

    res = proc(ctx)

    ctx.set_runs_on_server(False)
    ctx.config.set('use_server', True, config_name='config')
    return {'response': res, 'config': jsonpickle.encode(ctx.config)}

def __exception_message_with_all_causes(e):
    if isinstance(e, Exception) and e.__cause__:
        return str(e) + ' caused by ' + __exception_message_with_all_causes(e.__cause__)
    else:
        return str(e)

def __error_to_result(retval, einfo):
    res = __exception_message_with_all_causes(retval)

    if einfo:
        res += '\n' + str(einfo)

    return res

def __map_collection_to_name(res, collection_name):
    for provder in res.keys():
        if collection_name in res[provder]['data']:
            res[provder]['data'][collection_name] = list(
                map(lambda x: x.get('name'), res[provder]['data'][collection_name])
            )
